@@ -34,8 +34,8 @@ class FTWScanner
 {
    protected:
    vector<string> Patterns;
+   string Arch;
    const char *OriginalPath;
-   char *RealPath;
    bool ErrorPrinted;
    
    // Stuff for the delinker
@@ -69,8 +69,7 @@ class FTWScanner
    void AddPattern(string Pattern) { Patterns.push_back(Pattern); };
    bool SetExts(string Vals);
       
-   FTWScanner();
-   virtual ~FTWScanner() {delete [] RealPath;};
+   FTWScanner(string const &Arch = string());
 };
 
 class PackagesWriter : public FTWScanner
@@ -86,13 +85,13 @@ class PackagesWriter : public FTWScanner
    bool DoSHA256;
    bool NoOverride;
    bool DoContents;
+   bool LongDescription;
 
    // General options
    string PathPrefix;
    string DirStrip;
    FILE *Output;
    struct CacheDB::Stats &Stats;
-   string Arch;
 
    inline bool ReadOverride(string File) {return Over.ReadOverride(File);};
    inline bool ReadExtraOverride(string File) 
@@ -125,7 +124,7 @@ class ContentsWriter : public FTWScanner
    void Finish() {Gen.Print(Output);};
    inline bool ReadyDB(string DB) {return Db.ReadyDB(DB);};
    
-   ContentsWriter(string DB);
+   ContentsWriter(string DB, string const &Arch = string());
    virtual ~ContentsWriter() {};
 };
 
